@@ -32,6 +32,14 @@
  * Copyright (c) 1986-1991 by Sun Microsystems Inc.
  */
 
+#ifdef __APPLE__
+/*
+ * This #define causes netinet/in.h to pull in definitions for IPV6_PKTINFO and
+ * IPV6_RECVPKTINFO
+ */
+#define __APPLE_USE_RFC_3542
+#endif
+
 #include "config.h"
 
 /*
@@ -52,6 +60,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <netconfig.h>
+#include <netinet/in.h>
 #include <err.h>
 
 #include "rpc_com.h"
@@ -479,6 +488,9 @@ svc_dg_destroy_task(struct work_pool_entry *wpe)
 	    && rec->xprt.xp_fd != RPC_ANYFD) {
 		(void)close(rec->xprt.xp_fd);
 		rec->xprt.xp_fd = RPC_ANYFD;
+		if (rec->xprt.xp_fd_send != RPC_ANYFD)
+			close(rec->xprt.xp_fd_send);
+		rec->xprt.xp_fd_send = RPC_ANYFD;
 	}
 
 	if (rec->xprt.xp_ops->xp_free_user_data)
@@ -542,6 +554,16 @@ svc_dg_control(SVCXPRT *xprt, const u_int rq, void *in)
 	case SVCSET_XP_FLAGS:
 		xprt->xp_flags = *(u_int *) in;
 		break;
+	case SVCGET_XP_UNREF_USER_DATA:
+		mutex_lock(&ops_lock);
+		*(svc_xprt_void_fun_t *) in = xprt->xp_ops->xp_unref_user_data;
+		mutex_unlock(&ops_lock);
+		break;
+	case SVCSET_XP_UNREF_USER_DATA:
+		mutex_lock(&ops_lock);
+		xprt->xp_ops->xp_unref_user_data = *(svc_xprt_void_fun_t) in;
+		mutex_unlock(&ops_lock);
+		break;
 	case SVCGET_XP_FREE_USER_DATA:
 		mutex_lock(&ops_lock);
 		*(svc_xprt_fun_t *) in = xprt->xp_ops->xp_free_user_data;
@@ -576,6 +598,7 @@ svc_dg_override_ops(SVCXPRT *xprt, SVCXPRT *rendezvous)
 		ops.xp_reply = svc_dg_reply;
 		ops.xp_checksum = svc_dg_checksum;
 		ops.xp_unlink = svc_dg_unlink_it;
+		ops.xp_unref_user_data = NULL;	/* no default */
 		ops.xp_destroy = svc_dg_destroy_it;
 		ops.xp_control = svc_dg_control;
 		ops.xp_free_user_data = NULL;	/* no default */
@@ -601,6 +624,7 @@ svc_dg_rendezvous_ops(SVCXPRT *xprt)
 		ops.xp_reply = (svc_req_fun_t)abort;
 		ops.xp_checksum = NULL;		/* not used */
 		ops.xp_unlink = svc_dg_unlink_it;
+		ops.xp_unref_user_data = NULL;	/* no default */
 		ops.xp_destroy = svc_dg_destroy_it;
 		ops.xp_control = svc_dg_control;
 		ops.xp_free_user_data = NULL;	/* no default */
@@ -619,21 +643,15 @@ svc_dg_enable_pktinfo(int fd, const struct __rpc_sockinfo *si)
 
 	switch (si->si_af) {
 	case AF_INET:
-#ifdef SOL_IP
 		(void)setsockopt(fd, SOL_IP, IP_PKTINFO, &on, sizeof(on));
-#endif
 		break;
 
 	case AF_INET6:
-#ifdef SOL_IP
 		(void)setsockopt(fd, SOL_IP, IP_PKTINFO, &on, sizeof(on));
-#endif
-#ifdef SOL_IPV6
 		(void)setsockopt(fd, SOL_IPV6, IPV6_RECVPKTINFO,
 				 &on, sizeof(on));
 		(void)setsockopt(fd, SOL_IPV6, IPV6_V6ONLY,
 				 &off, sizeof(off));
-#endif
 		break;
 	}
 }
@@ -689,19 +707,15 @@ svc_dg_store_pktinfo(struct msghdr *msg, SVCXPRT *xprt)
 	     cmsg = CMSG_NXTHDR(msg, cmsg)) {
 
 		if (cmsg->cmsg_level == IPPROTO_IP) {
-#ifdef IP_PKTINFO
 			if (cmsg->cmsg_type == IP_PKTINFO) {
 				if (svc_dg_store_in_pktinfo(cmsg, xprt))
 					return 1;
 			}
-#endif
 
-#ifdef IPV6_PKTINFO
 			if (cmsg->cmsg_type == IPV6_PKTINFO) {
 				if (svc_dg_store_in6_pktinfo(cmsg, xprt))
 					return 1;
 			}
-#endif
 		}
 		
 	}
